#==============================================================================
# C O P Y R I G H T
#------------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
#==============================================================================

import os

# Public API:
# Cache
from configos.cache import CACHE_LARGE
from configos.cache import CACHE_MEDIUM
from configos.cache import CACHE_SMALL
from configos.cache import cache_large
from configos.cache import cache_medium
from configos.cache import cache_small
# debug
from configos.debugs import debug
from configos.debugs import debug_set
from configos.debugs import debug_unset
# directory
from configos.directory import check_startup
from configos.directory import environment
from configos.directory import export
from configos.directory import makedirs
from configos.directory import ready
from configos.directory import share
from configos.directory import tmp
from configos.directory import todo
# docs
from configos.docs import docs_url
# Document
from configos.document import OneSideDINA4
from configos.document import OneSideDINA5
# env
from configos.env import env
from configos.env import env_del
from configos.env import env_dump
from configos.env import env_load
from configos.env import env_path_append
from configos.env import env_path_remove
from configos.env import env_set
from configos.env import env_unload
# Exception
from configos.exception import HolyValueError
from configos.exception import InvalidHolyValue
from configos.exception import MissingHolyValue
# Holy
from configos.holyvalue.access import HV
from configos.holyvalue.access import HV_API
from configos.holyvalue.access import HV_BOOL
from configos.holyvalue.access import HV_FLOAT
from configos.holyvalue.access import HV_FLOAT_MINUS
from configos.holyvalue.access import HV_FLOAT_PLUS
from configos.holyvalue.access import HV_GB
from configos.holyvalue.access import HV_HOUR
from configos.holyvalue.access import HV_INT
from configos.holyvalue.access import HV_INT_MINUS
from configos.holyvalue.access import HV_INT_PLUS
from configos.holyvalue.access import HV_KB
from configos.holyvalue.access import HV_MB
from configos.holyvalue.access import HV_MINUTE
from configos.holyvalue.access import HV_PERCENT
from configos.holyvalue.access import HV_PERCENT_MINUS
from configos.holyvalue.access import HV_PERCENT_PLUS
from configos.holyvalue.access import HV_SECOND
from configos.holyvalue.access import HV_SECRET
from configos.holyvalue.access import HV_STR
from configos.holyvalue.cloud import cloud_base
from configos.holyvalue.cloud import cloud_base_set
from configos.holyvalue.cloud import cloud_lookup
from configos.holyvalue.cloud import cloud_set
from configos.holyvalue.cloud import cloud_unset
from configos.holyvalue.cloud import holyname
from configos.holyvalue.collect import generate
from configos.holyvalue.data import NOMATH
from configos.holyvalue.data import DataType
from configos.holyvalue.data import HolyValue
from configos.holyvalue.store import database
from configos.holyvalue.store import init
from configos.holyvalue.store import load
from configos.holyvalue.store import parse
from configos.holyvalue.table import HolyList
from configos.holyvalue.table import HolyRate
from configos.holyvalue.table import HolyTable
# Server
from configos.server import package_address
from configos.server import package_configuration

__version__ = '1.0.1'

ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
PROCESS = 'configos'

init(cloud_base())

# TODO: REMOVE LATER AND INCREASE MAJOR VERSION NUMBER
load_env = env_load
unload_env = env_unload
dump_env = env_dump



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import inspect
import os

import pytest
import utilo
import utilotest

import configos
import configos.holyvalue
import configos.holyvalue.data
import configos.holyvalue.store
import tests

EXAMPLE = os.path.join(tests.TEST_DATA, 'examples')
FIRST_ONE = 'first_one'


@pytest.fixture
def default_one():
    """Init `DataBase` with `FIRST_ONE` example"""
    configos.init(EXAMPLE)
    configos.load(FIRST_ONE)


def test_holyvalue_load_databas():
    parsed = configos.holyvalue.store.parse(
        os.path.join(EXAMPLE, f'{FIRST_ONE}.hv'),
        FIRST_ONE,
    )
    assert parsed
    assert len(parsed.data) == 2


@pytest.mark.usefixtures('default_one')
def test_holyvalue_hv():
    distance = configos.HV(group='groupme.footer.header', name='DISTANCE')
    # No convertion is done, cause of not defining `datatype`
    assert distance == '50022', distance


@pytest.mark.usefixtures('default_one')
def test_holyvalue_hv_group_from_module(mp):
    """Test to determine variable out of module and variable assignment"""

    def getmodule(__=None, _=None):  # pylint:disable=W0613
        # patch object.__name__
        return type('groupme.footer.header', (tuple,), {})

    with mp.context() as context:
        context.setattr(inspect, 'getmodule', getmodule)
        distance = configos.HV(
            datatype=configos.holyvalue.data.DataType.INT_PLUS,
            limit=60000,
        )

    assert distance == 50022, distance


def test_holyvalue_hv_use_default(capsys):
    """Test returning `default` value for non defined variables and
    inform developer about this."""
    default = 'Helm'
    with utilo.level_tmp(utilo.Level.DEBUG):
        result = configos.HV(default=default).value
    assert result == default
    # ensure to log warning
    stdout = utilotest.stdout(capsys)
    assert 'not defined' in stdout, stdout


def test_holyvalue_hv_use_no_default():
    with pytest.raises(configos.MissingHolyValue):
        _ = configos.HV().value


def test_holyvalue_invalid_variable():
    """Test that variable not exists"""

    with pytest.raises(configos.MissingHolyValue):
        _ = configos.HV(
            group='groupme.footer.header',
            name='does_not_exists',
        ).value

    # use default value instead of throwing MissingHolyValue-Exception
    default = 100
    got = configos.HV(
        group='groupme.footer.header',
        name='does_not_exists',
        default=100,
    )
    assert got == default


@pytest.mark.usefixtures('default_one')
@pytest.mark.parametrize('datatype', configos.NOMATH)
def test_holyvalue_invalid_limit(datatype):
    """Test to determine variable out of module and variable assignment"""
    with pytest.raises(configos.InvalidHolyValue):
        _ = configos.HV(
            group='groupme.footer.header',
            name='distance',
            datatype=datatype,
            limit=100,  # LIMIT IS HIGHER THAN DEFINED 50022 in first_one.hv
        ).value


def test_holyvalue_generate_configuration():
    """If this test fails, check holyvalue() signature and config generator"""
    config = configos.generate(tests.HVEXAMPLE)
    assert config is not None
    assert len(config) > 200, 'no enough content'

    keys = ['FIRST', 'SECOND', 'THIRD', 'LEVEL_UP', 'HELMUT']
    for key in keys:
        assert f'{key} = ' in config, utilo.log_raw(config)

    variables = ['default', 'limit', 'datatype', 'comment']
    for variable in variables:
        assert f'# {variable}:' in config, print(config)

    assert config.count('#') >= 5, utilo.log_raw(config)


def test_holyvalue_generate_and_load(td):
    path = td.tmpdir.join('config.hv')
    # create config
    config = configos.generate(tests.HVEXAMPLE)
    utilo.file_create(path, config)
    # parse config
    parsed = configos.holyvalue.store.parse(path, 'config')
    assert parsed


def test_holyvalue_evaluate_percent_plus():
    with pytest.raises(configos.InvalidHolyValue):
        _ = configos.HV(
            default=15,
            limit=10,
            datatype=configos.DataType.PERCENT_PLUS,
        )


def test_holyvalue_less_verbose_api():
    access = configos.HV_INT_PLUS(default=5)
    assert access == 5


def test_holyvalue_default_database():
    hello = configos.HV_INT_PLUS(default=5)
    assert hello == 5


def test_holyvalue_right_hand_evaluation_name_and_group():
    abc = configos.HV(
        name='alpha',
        default=15,
        limit=120,
        datatype=configos.DataType.PERCENT_PLUS,
    )
    assert abc.name == 'alpha'
    assert abc.group == 'tests.test_holyvalue'


def test_holyvalue_operation():
    smaller = configos.HV(
        default=5,
        name='smaller',
        datatype=configos.DataType.INT,
    )
    value = configos.HV(
        default=15,
        name='alpha',
        datatype=configos.DataType.INT,
    )
    string = configos.HV(
        default='string',
        name='alpha',
        datatype=configos.DataType.STR,
    )
    assert value + 15 == 30
    assert value + value == 30

    assert value - 15 == 0  # pylint:disable=C2001
    assert value - value == 0  # pylint:disable=C2001

    assert value * 15 == 15 * 15
    assert value * value == 15 * 15

    assert value / 15 == 1
    assert value / value == 1

    assert value % 2 == 1
    assert value % value == 0  # pylint:disable=C2001

    assert value >= (value * 0.5)
    assert smaller <= value  # pylint:disable=comparison-with-itself

    assert value == value  # pylint:disable=comparison-with-itself

    assert value > (value * 0.5)
    assert smaller < (value * 2)

    assert (value * 2) > value
    assert (value * 2) >= value

    assert value != (value * 2)
    assert value
    assert value is value  # pylint:disable=comparison-with-itself

    assert string == string  # pylint:disable=comparison-with-itself


def test_hv_ranged():
    start = configos.HV_INT_PLUS(default=10)
    end = configos.HV_INT_PLUS(default=20)
    assert len(utilo.rtuple(start, end)) == end - start


def test_hv_slice():
    data = [1, 2, 3, 4, 5, 6, 7]
    start = configos.HV_INT_PLUS(default=2)
    selected = data[start:]
    assert selected == [3, 4, 5, 6, 7]



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================
"""\
>>> STR = HV_STR('hello')
>>> f'{STR} hier'
'hello hier'
"""

import functools
import inspect

import utilo

import configos
import configos.holyvalue
import configos.holyvalue.collect
import configos.holyvalue.data

NO_GROUP = 'NO_GROUP'


def holyvalue(
    default=None,
    *,
    name: str = None,
    group: str = None,
    limit=None,
    datatype: 'DataType' = None,
) -> 'HolyValue':
    """Access `holyvalue` via `variable` and `group`.

    Args:
        default(DataType): define default variable for non defined
                           variable to avoid updating configration to
                           often.
        name(str): name of holy value
        group(str): module where `var` is located
        limit(DataType): limit to validate configuration file
        datatype(DataType): convert str-based configuration file
    Raises:
        InvalidHolyValue: if value of configuration file hits requirements
        MissingHolyValue: if value is not defined and no default one is defined
    Returns:
        Defined HolyValue in `config_name.hv` or `default` one.

    TODO: MAKE method update able/facade/callable
    """
    assert name is None or isinstance(name, str), f'invalid name {name}'
    assert group is None or isinstance(group, str), f'invalid group: {group}'
    if name is None:
        # TODO: REMOVE THIS HACK
        # TODO: NOT VERY STABLE/ DIRTY
        # determine variable out of code
        levelup = inspect.stack(context=1)[1].code_context
        code = utilo.NEWLINE.join(levelup)
        matched = utilo.search(configos.holyvalue.collect.PATTERN, code)
        name = str(matched['variable']).strip().upper()
    if group is None:
        # determine call package
        frame = inspect.currentframe()
        parent = frame.f_back  # get invoker
        inspected = inspect.getmodule(parent)
        if inspected:
            group = inspect.getmodule(inspected).__name__
        else:
            # is loaded later via dynamic code loader
            utilo.debug(f'could not determine holyvalue group: {parent}')
            group = NO_GROUP
    if not configos.holyvalue.data.validate(default, datatype, default, limit):
        msg = f'invalid default/limit/datatype; name: {name}; group:{group} '
        msg += f'default: {default}; limit: {limit}; type: {datatype}'
        raise configos.InvalidHolyValue(msg)
    result = configos.HolyValue(name, group, datatype, default, limit)
    return result


DataType = configos.holyvalue.data.DataType
init = lambda datatype: functools.partial(holyvalue, datatype=datatype)
HV = holyvalue  # pylint:disable=C0103

HV_BOOL = init(DataType.BOOL)  # pylint:disable=C0103
HV_STR = init(DataType.STR)  # pylint:disable=C0103

HV_INT = init(DataType.INT)  # pylint:disable=C0103
HV_INT_PLUS = init(DataType.INT_PLUS)  # pylint:disable=C0103
HV_INT_MINUS = init(DataType.INT_MINUS)  # pylint:disable=C0103

HV_FLOAT = init(DataType.FLOAT)  # pylint:disable=C0103
HV_FLOAT_PLUS = init(DataType.FLOAT_PLUS)  # pylint:disable=C0103
HV_FLOAT_MINUS = init(DataType.FLOAT_MINUS)  # pylint:disable=C0103

HV_PERCENT = init(DataType.PERCENT)  # pylint:disable=C0103
HV_PERCENT_PLUS = init(DataType.PERCENT_PLUS)  # pylint:disable=C0103
HV_PERCENT_MINUS = init(DataType.PERCENT_MINUS)  # pylint:disable=C0103

HV_SECOND = init(DataType.SECOND)  # pylint:disable=C0103
HV_MINUTE = init(DataType.MINUTE)  # pylint:disable=C0103
HV_HOUR = init(DataType.HOUR)  # pylint:disable=C0103

HV_API = init(DataType.API)  # pylint:disable=C0103
HV_SECRET = init(DataType.SECRET)  # pylint:disable=C0103

HV_KB = init(DataType.KB)  # pylint:disable=C0103
HV_MB = init(DataType.MB)  # pylint:disable=C0103
HV_GB = init(DataType.GB)  # pylint:disable=C0103



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import contextlib
import dataclasses
import enum
import functools

import utilo

import configos


class HolyMixin:
    pass


class DataType(enum.Enum):
    INT = enum.auto()
    INT_PLUS = enum.auto()
    INT_MINUS = enum.auto()

    FLOAT = enum.auto()
    FLOAT_PLUS = enum.auto()
    FLOAT_MINUS = enum.auto()

    PERCENT = enum.auto()
    PERCENT_PLUS = enum.auto()
    PERCENT_MINUS = enum.auto()

    BOOL = enum.auto()
    STR = enum.auto()

    SECOND = enum.auto()
    MINUTE = enum.auto()
    HOUR = enum.auto()

    API = enum.auto()
    SECRET = enum.auto()

    KB = enum.auto()
    MB = enum.auto()
    GB = enum.auto()


TIME = {DataType.SECOND, DataType.MINUTE, DataType.HOUR}
BYTES = {DataType.KB, DataType.MB, DataType.GB}

NOMATH = {
    item for item in DataType if item not in (
        DataType.API,
        DataType.BOOL,
        DataType.SECRET,
        DataType.STR,
    )
}


@dataclasses.dataclass
class Datum:
    name: str = None
    value: object = None
    description: str = None
    datatype: DataType = None
    limit = None
    unit: str = None


@dataclasses.dataclass
class Group:
    """\
    >>> Group()
    Group(name=None, data={})
    """
    name: str = None
    data: dict[str, Datum] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class DataSet:
    name: str = None
    data: dict[str, Group] = dataclasses.field(default_factory=dict)


@functools.lru_cache(maxsize=1024)
def convert(data, datatype=None):  # pylint:disable=R1260
    """\
    >>> convert(1, DataType.BOOL)
    True
    >>> convert(70, DataType.PERCENT_PLUS)
    0.7
    """
    if datatype is None:
        return data
    # convert value
    datatype = str(datatype)
    if 'INT' in datatype:
        data = int(data)
    elif 'FLOAT' in datatype:
        data = float(data)
    elif 'PERCENT' in datatype:
        data = float(data)
        data *= 0.01
        data = utilo.roundme(
            data,
            digits=3,
        )
    elif 'BOOL' in datatype:
        data = utilo.str2bool(data)
    elif 'SECRET' in datatype and isinstance(data, str):
        data: bytes = data.encode('ascii')
    elif 'KB' in datatype:
        data = int(float(data) * 1024)
    elif 'MB' in datatype:
        data = int(float(data) * 1024 * 1024)
    elif 'GB' in datatype:
        data = int(float(data) * 1024 * 1024 * 1024)
    elif any(item in datatype for item in ('SECOND', 'MINUTE', 'HOUR')):
        data = int(data)
    return data


class HolyValue(HolyMixin):
    """\
    >>> 10.0 + HolyValue(default=5.0)
    15.0
    >>> 10.0 - HolyValue(default=5.0)
    5.0
    >>> 10 / HolyValue(default=5.0)
    2.0
    >>> HolyValue(default=5.0) / 10
    0.5
    """

    def __init__(self, name=None, group=None, datatype=None, default=None, limit=None):  # yapf:disable
        self.hvname = name
        self.hvgroup = group
        self.datatype = datatype
        self.default = default
        self.limit = limit

    @functools.cached_property
    def value(self):
        assert configos.database(), 'could not access database'
        if not self.valid:
            msg = (f'invalid holyvalue: {self.group}:{self.name};\n'
                   f'value:{self.data}; default:{self.default};\n'
                   f'limit:{self.limit}; datatype:{self.datatype}')
            raise configos.exception.InvalidHolyValue(msg)
        # TODO: USE BETTER CACHING MECHANISM
        return self.data

    @property
    def valid(self):
        return validate(
            self.data,
            datatype=self.datatype,
            default=self.default,
            limit=self.limit,
        )

    @property
    def name(self):
        return self.hvname

    @property
    def group(self):
        return self.hvgroup

    @property
    def data(self):
        result = configos.database().get(
            group=self.group,
            variable=self.name,
            default=self.default,
        )
        result = convert(result, datatype=self.datatype)
        return result

    def __add__(self, other):
        with contextlib.suppress(TypeError):
            return self.value + other
        return self.value + other.value

    __radd__ = __add__

    def __sub__(self, other):
        with contextlib.suppress(TypeError):
            return self.value - other
        return self.value - other.value

    def __rsub__(self, other):
        return other - self.value

    def __mul__(self, other):
        with contextlib.suppress(TypeError):
            return self.value * other
        return self.value * other.value

    __rmul__ = __mul__

    def __truediv__(self, other):
        with contextlib.suppress(TypeError):
            return self.value / other
        return self.value / other.value

    def __rtruediv__(self, other):
        return other / self.value

    def __mod__(self, other):
        with contextlib.suppress(TypeError):
            return self.value % other
        return self.value % other.value

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other

    def __eq__(self, other):
        return self.value == other

    def __int__(self):
        return int(self.value)

    def __bool__(self):
        """\
        >>> bool(HolyValue(default=False))
        False
        >>> HolyValue(default=False) == False
        True
        >>> assert not HolyValue(default=False)
        >>> assert HolyValue(default=True)
        """
        return bool(self.value)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        """\
        >>> str((HolyValue(default=5.0), HolyValue(default=7.1)))
        '(5.0, 7.1)'
        """
        return str(self)

    def __index__(self):
        return int(self)

    def __hash__(self):
        """\
        >>> assert hash(HolyValue(default=5.0))
        """
        return hash(repr(self))


def validate(  # pylint:disable=R0911,R1260
    data,
    datatype: DataType = None,
    default=None,
    limit=None,
) -> bool:
    """\
    >>> validate(True, DataType.BOOL, True)
    True
    >>> validate(51, DataType.INT, limit=50)
    False
    >>> validate(-10, DataType.INT_PLUS)
    False
    >>> validate(10.0, DataType.INT_PLUS)
    False
    >>> validate(10.0, DataType.FLOAT_MINUS)
    False
    """
    with contextlib.suppress(TypeError):
        # avoid that default is higher than limit
        if default > limit:
            return False
    if data is None:
        return True
    if datatype:
        if 'INT' in datatype.name or datatype in TIME:
            if not utilo.isint(data):
                return False
        if 'PLUS' in datatype.name or datatype in TIME or datatype in BYTES:
            if data < 0.0:
                return False
        if 'MINUS' in datatype.name:
            if data > 0.0:
                return False
    if limit is not None and datatype != DataType.BOOL:
        if data > limit:
            return False
    return True



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import configparser
import functools
import os

import utilo

import configos.exception
import configos.holyvalue
import configos.holyvalue.data

DATABASE = None


class DataBase:
    """Define facade to access variable via `group` and `variable` name."""

    def __init__(self, path: str, current: str = None):
        """Setup `DataBase` with no `current` DataSet. Use `load` or
        `current` to init `DataSet`.

        Args:
            path(str): folder with different `DataSet`s
            current(str): name of default dataset
        """
        if not current:
            current = configos.holyvalue.data.DataSet()
        self.path = path
        self.current = current

    def load(self, name: str, path: str = None):
        path = path if path else self.path
        path = os.path.join(path, f'{name}.hv')
        utilo.info(f'load path: {path}')
        parsed = parse(path, name)
        self.current = parsed

    @functools.lru_cache(maxsize=1024)
    def get(self, group: str, variable: str, default=None):
        if group:
            group = group.lower()
        if variable:
            variable = variable.upper()

        def default_warning():
            msg = f'not defined in database {group}:{variable}; use default: {default}'
            utilo.info(msg)

        # determine hv-group
        _group = self.current.data.get(group, None)  # pylint:disable=E1101
        if not _group:
            if default is not None:
                default_warning()
                return default
            raise configos.exception.MissingHolyValue(f'invalid group: {group}')
        # determine hv-name
        variable = variable.upper()
        _var = _group.data.get(variable.upper(), None)
        if not _var:
            if default is not None:
                default_warning()
                return default
            msg = f'invalid variable: {group}:{variable}'
            raise configos.exception.MissingHolyValue(msg)
        return _var.value


def parse(path, name: str = None) -> 'DataSet':
    # load config
    raw = utilo.from_raw_or_path(path, ftype='ini')
    parser = configparser.ConfigParser()
    parser.read_string(raw)
    # prepare data
    dataset = configos.holyvalue.data.DataSet(name=name)
    for groupname, groupdata in parser.items():
        if groupname == 'DEFAULT':
            continue
        groupname = groupname.lower()
        group = configos.holyvalue.data.Group(name=groupname)
        for datakey, data in groupdata.items():
            datakey = datakey.upper()
            datum = configos.holyvalue.data.Datum(name=datakey, value=data)
            group.data[datakey] = datum  # pylint:disable=E1137
        dataset.data[groupname] = group  # pylint:disable=E1137
    return dataset


def init(path: str):
    """Init `DATABASE` with configuration `path`."""
    assert path is None or os.path.exists(path), f'path does not exists: {path}'
    global DATABASE  # pylint:disable=global-statement
    DATABASE = DataBase(path)


def load(name: str, base: str = None):
    """Load `DATABASE` with dataset of `name`."""
    database().load(name=name, path=base)


def database():
    """Access global variable `DATABASE`"""
    return DATABASE



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2020-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================
"""HolyTable lookup
================

>>> table = HolyTable(strategy=utilo.Strategy.LINEARISE)
>>> table.add(0,0)
>>> table.add(10,10)
>>> assert table(5.0) == 5.0, table(5.0)

>>> rate = HolyRate(((1, 1), (2, 2), (5, 1), (10, 5)))
>>> assert rate(2) == 1.0
>>> assert rate(6) == 0.2
>>> assert rate(10) == 0.2

Support both lookup styles
>>> assert rate(10) == rate[10]

"""

import utilo

import configos.holyvalue.data


class HolyTable(configos.holyvalue.data.HolyMixin):

    def __init__(
        self,
        items: list = None,
        strategy: utilo.Strategy = None,
        right_outranges_none: bool = False,
        left_outranges_none: bool = False,
    ):
        """HolyTable Looup

        Args:
            items: data points to construct lookup values
            strategy: strategy to construct lookup values
            left_outranges_none: return None if request value is out of
                                 defined range
            right_outranges_none: return None if request value is out of
                                 defined range
        """
        self.table = []
        self.strategy = strategy
        self.right_outranges_none = right_outranges_none
        self.left_outranges_none = left_outranges_none
        if items:
            for item in items:
                self.add(*item)

    def add(self, position, value):
        self.table.append((position, value))
        self.table = sorted(self.table, key=lambda x: x[0])
        assert utilo.isascending([item[0] for item in self.table]), str(self.table) # yapf:disable

    def __call__(self, position):
        """Return lookup value or None.

        None: If request point is out of defined area and no fallback is
              activated.
        """
        assert self.table, 'empty table, use add to configure'
        value = utilo.lookup(
            position,
            self.table,
            self.strategy,
            self.right_outranges_none,
            self.left_outranges_none,
        )
        return value

    def __getitem__(self, index):
        return self(index)


class HolyRate(HolyTable):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.table = [
            (pos, value / pos if pos else 0.0) for pos, value in self.table
        ]


class HolyList(configos.holyvalue.data.HolyMixin):
    """A HolyList is a list which can be configured by HolyValue-Mechanism.

    Therefore a live-configuration is possible.

    >>> data = HolyList((1, 3, 5))
    >>> assert list(data) == [1, 3, 5]
    """

    def __init__(self, items):
        self.items = items

    def __getitem__(self, index):
        return self.items[index]



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import os

from pytest import raises

from configos import package_address
from configos import package_configuration
from configos.server import HELPY_EXT_DIRECT
from configos.server import HELPY_EXT_PORT
from configos.server import HELPY_INT_DIRECT
from configos.server import HELPY_INT_PORT
from configos.server import HELPY_URL


def test_missing_environment(mp):
    with mp.context() as context:
        # Remove all environment vars
        context.setattr(os, 'environ', {})

        with raises(SystemExit):
            package_address()

        with raises(SystemExit):
            package_configuration()


def test_package_adress_and_configuration(mp):
    with mp.context() as context:
        # Remove all environment vars
        context.setattr(
            os, 'environ', {
                HELPY_URL: 'http://packages.checkitweg.de',
                HELPY_INT_PORT: 8081,
                HELPY_EXT_PORT: 8082,
                HELPY_INT_DIRECT: HELPY_URL + ':' + HELPY_INT_PORT,
                HELPY_EXT_DIRECT: HELPY_URL + ':' + HELPY_EXT_PORT,
            })

        config = package_configuration()
        assert len(config) == 3

        adress = package_address()
        assert len(adress) == 2



#==============================================================================
# C O P Y R I G H T
#------------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
#==============================================================================

import os
import sys

import utilo

HELPY_URL = 'HELPY_URL'

HELPY_INT_PORT = 'HELPY_INT_PORT'
HELPY_EXT_PORT = 'HELPY_EXT_PORT'

HELPY_INT_DIRECT = 'HELPY_INT_DIRECT'
HELPY_EXT_DIRECT = 'HELPY_EXT_DIRECT'


def package_configuration():
    """Return tuple with adress, internal package port and external package
    port of the package repository"""

    try:
        adress = os.environ[HELPY_URL]
        internal = os.environ[HELPY_INT_PORT]
        external = os.environ[HELPY_EXT_PORT]
        internal, external = int(internal), int(external)
        return (adress, internal, external)
    except KeyError as msg:
        handle_error(msg)
    return None


def package_address():
    """Return tuple of direct addresses to internal and external package
    repository"""

    try:
        internal = os.environ[HELPY_INT_DIRECT]
        external = os.environ[HELPY_EXT_DIRECT]
        return (internal, external)
    except KeyError as msg:
        handle_error(msg)
    return None


def handle_error(msg: KeyError):
    utilo.error('Missing global var: %s' % msg)
    sys.exit(utilo.FAILURE)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import io
import os
import re
import tokenize

import utilo

import configos.holyvalue.access
import configos.holyvalue.data
import configos.utils


def generate(path: str, skips: callable = None) -> str:
    """Iterate thrue python files and extract config with holyvalues
    constructs out of project structure.

    Args:
        path(str): project root path
        skips(callable): should we skip file
    Returns:
        project configuration file
    """
    assert os.path.exists(path), f'path does not exists: {path}'
    collected = collect(path, skips=skips)
    root = rootpackage(path)
    result = dump_collected(collected, root)
    return result


def collect(path: str, skips: callable = None) -> dict:
    path = os.path.abspath(path)
    files = utilo.file_list(
        path,
        include='py',
        absolute=True,
        recursive=True,
    )
    collected = {}
    for item in files:
        if skips and skips(item):
            continue
        relative = utilo.make_relative(item, path)
        relative = utilo.make_package(relative)
        parsed = holyvalue_from_file(item)
        if parsed:
            collected[relative] = parsed
    return collected


def dump_collected(collected, root):
    signature = utilo.attributes(configos.holyvalue.access.holyvalue)
    raw = []
    for package in sorted(collected.keys()):
        raw.append(f'[{root}.{package}]')
        for variable, values in collected[package].items():
            for item, value in values.items():
                if item == 'table':
                    continue
                raw.append(f'# {item}:{value}')
            assert 'name' in signature, 'require name'
            variable = values.get('hvname', variable)
            variable = variable.replace("'", '').upper()
            assert 'default' in signature, 'require default'
            default = values.get('default', 'None')
            if default == 'None':
                default = values.get('table', 'None')
            raw.append(f"{variable} = {default}")
            raw.append('')
        raw.append('')
    result = utilo.NEWLINE.join(raw)
    result = result.strip()
    return result


def rootpackage(root: str) -> str:
    package = os.path.split(root)[1]
    # rawmaker-2.26.6-py3.8.egg
    package = str(package).split('-', maxsplit=1)[0]
    return package


def holyvalue_from_file(path: str) -> dict:
    """Parse holyvalues from `sourcecode`.

    Args:
        path(str): python source code file
    Returns:
        Dictionary with holyvalues and further configuration parameter,
        e.g. limit, variable, group etc.
    """
    commento = comments(utilo.file_read(path))
    module = configos.utils.load_module(path)
    if not module:
        utilo.error(f'could not load: {path}')
        return {}
    result = {}
    for key, value in vars(module).items():
        if not isinstance(value, configos.holyvalue.data.HolyMixin):
            continue
        result[key] = {'comment': commento[key], **vars(value)}
    return result


PATTERN = r"""
    \b
    (?P<variable>[\w\d_]+)[ ]=[ ]
    (
        configos\.HV[\w\d_]*?|
        configos\.HolyTable|
    )
"""


def comments(sourcecode: str) -> dict:
    lines = codelines(sourcecode)
    result = {}
    for line, comment in lines.items():
        matched = re.match(PATTERN, line, re.MULTILINE | re.VERBOSE)
        if not matched:
            continue
        variable = matched['variable']
        result[variable] = comment
    return result


def codelines(sourcecode: str):
    """Remove comments etc. out of `sourcecode`."""
    lines = {}
    tokenized = list(token(sourcecode))
    for index, item in enumerate(tokenized[1:], start=1):
        if item.type != tokenize.NAME:
            # skip comments etc.
            continue
        stripped = item.line.strip()
        holycomment = ''
        # TODO: SUPPORT MULTILINE COMMENT
        # look one line back to check if holy value has a holy comment
        if tokenized[index - 1].type in (tokenize.NL,):  # comment
            holycomment = tokenized[index - 1].line
            holycomment = holycomment.replace('#', '', 1)
            holycomment = holycomment.strip()
        if lines.get(stripped, None):
            # do not overwrite holy comments
            continue
        lines[stripped] = holycomment
    return lines


def token(code: str):
    buffer = io.BytesIO(code.encode(utilo.UTF8)).readline
    source = tokenize.tokenize(buffer)
    return source



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import collections
import csv
import os

import utilo

import configos


def show(result):
    single = result[0]
    parsed, header, size = parse_result(single)
    rendered = render_table(list(parsed.values()), size, header=header)
    with utilo.make_tmpdir(root=configos.ROOT) as output:
        outpath = os.path.join(output, 'index.html')
        utilo.file_create(outpath, rendered)
        testrun = utilo.testing()
        if not testrun:
            utilo.run(f'start {outpath}')


def parse_result(path) -> dict:
    collected = collections.defaultdict(set)
    with open(path, newline='', encoding='utf8') as csvfile:
        reader = csv.reader(csvfile)
        data = list(reader)
    width, height = 0, 0
    header = data[0][1:-1]
    for row in data[1:]:
        number, *content, failure = row  # pylint:disable=W0612
        failure = int(failure)
        for index, value in enumerate(content):
            # TODO: ADD BOOL CHECKUP
            value = int(value) if utilo.isint(value) else float(value)
            collected[index].add((value, failure))
        width = len(content)
        height += 1
    result = {key: prepare(value) for key, value in collected.items()}
    return result, header, (width, height)


def prepare(value):
    grouped = utilo.groupby_x(value, selector=lambda x: x[0])
    result = []
    for group in grouped:
        result.append((group[0][0], min((item[1] for item in group))))
    return sorted(result, key=lambda x: x[0], reverse=True)


def render_table(data, size, header=None) -> str:
    width = 150 * size[0]
    result = f'<html><table width={width}>'
    if header:
        result += '<tr>'
        result += ''.join([
            f'<td style="min-width:150px;font-size:10px;overflow:hidden">{item.replace(".", " ")}</td>'
            for item in header
        ])
        result += '</tr>'
    for _ in range(size[1]):  # pylint:disable=W0612
        result += '<tr>'
        for pos in range(size[0]):
            current = data[pos]
            if current:
                value, failure = current.pop()
                color = 'green' if not failure else 'orange'
                result += f'<td style="background:{color};" height=50px><center>{value} [{failure}]</center></td>'
            else:
                result += '<td></td>'
        result += '</tr>'
    result += '</table></html>'
    return result



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import configos


def test_debug_set_unset():
    assert not configos.debug()
    configos.debug_set()
    assert configos.debug()
    configos.debug_unset()
    assert not configos.debug()



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import utilo

import configos

DEBUG = 'KIWI_DEBUG'


def debug() -> bool:
    try:
        current = configos.env(DEBUG)
    except KeyError:
        return False
    result = utilo.str2bool(current)
    return result


def debug_set():
    configos.env_set(DEBUG, 'True')


def debug_unset():
    configos.env_set(DEBUG, 'False')



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================
"""Access path's of global data share. These locations are stored in
global environment variables:

+ SHARED_READY
+ SHARED_SPACE
+ SHARED_TMP
+ SHARED_TODO

Use `configos.export()` to change these variables.
"""

import os
import sys

import utilo

TODO = 'SHARED_TODO'
READY = 'SHARED_READY'
COMMON = 'SHARED_SPACE'
TMP = 'SHARED_TMP'


def ready(check: bool = False) -> str:
    """Return path to finished jobs.

    Args:
        check(bool): check existence of path
    Returns:
        path from global variable
    Raises:
        SystemExit if check is True and path does not exists
        SystemExit if global env var does not exist
    """
    return _path_from_env(READY, check=check)


def todo(check: bool = False) -> str:
    """Return path todo with jobs to work.

    Args:
        check(bool): check existence of path
    Returns:
        path from global variable
    Raises:
        SystemExit if check is True and path does not exists
        SystemExit if global env var does not exist
    """
    return _path_from_env(TODO, check=check)


def share(check: bool = False) -> str:
    """Return path to shared folder.

    Args:
        check(bool): check existence of path
    Returns:
        path from global variable
    Raises:
        SystemExit if check is True and path does not exists
        SystemExit if global env var does not exist
    """
    return _path_from_env(COMMON, check=check)


def tmp(check: bool = False) -> str:
    """Return path to tmp folder.

    Args:
        check(bool): check existence of path
    Returns:
        path from global variable
    Raises:
        SystemExit if check is True and path does not exists
        SystemExit if global env var does not exist
    """
    return _path_from_env(TMP, check=check)


def export(common: str, todo: str, ready: str):  # pylint:disable=W0621
    """Set environment variables"""
    os.environ[COMMON] = str(common)
    os.environ[TODO] = str(todo)
    os.environ[READY] = str(ready)


def environment(check: bool = False):
    """Return `SHARED_TODO` and `SHARED_READY` folder."""
    todo_ = todo(check=check)
    ready_ = ready(check=check)
    common = share(check=check)

    return common, todo_, ready_


def _path_from_env(env: str, check: bool = False):  # pylint:disable=R1710
    """Access global env variable."""
    assert env
    try:
        path = os.environ[env]
        if check and not os.path.exists(path):
            utilo.error('Path does not exists: %s' % path)
            sys.exit(utilo.FAILURE)
        return path
    except KeyError:
        utilo.error('Missing environment var `%s`' % env)
        sys.exit(utilo.FAILURE)


def makedirs():
    """Ensure that all working directories are created."""
    os.makedirs(ready(), exist_ok=True)
    os.makedirs(share(), exist_ok=True)
    os.makedirs(tmp(), exist_ok=True)
    os.makedirs(todo(), exist_ok=True)


def check_startup():
    """Check that global environments are set."""
    ready(check=True)
    share(check=True)
    tmp(check=True)
    todo(check=True)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import utilo

import configos.cli.plan
import configos.cli.result


def evaluate(
    create: list,
    run: str,
    show: str,
    reduce: int,
    cmd_test: str,
):  # pylint:disable=W0613
    if create:
        plan = configos.cli.plan.create(create)
        utilo.log(configos.cli.plan.dump(plan))
    if run:
        if cmd_test is None:
            cmd_test = 'baw test -n1'
        configos.cli.plan.run(
            run,
            reduce=reduce,
            cmd_test=cmd_test,
        )
    if show:
        configos.cli.result.show(show)


def add_option(parser):
    # TODO: REPLACE WITH UTILA METHOD
    sub = parser.add_subparsers(help='run optimizer to determine holy values')
    show = sub.add_parser('optimize')
    show.add_argument(
        '--create',
        help='create optimization plan',
        action='append',
    )
    show.add_argument(
        '--run',
        help='run optimization',
        action='append',
    )
    show.add_argument(
        '--show',
        help='show optimization result',
        action='append',
    )
    show.add_argument(
        '-r',
        default=100,
        type=int,
        help='number of optimization steps',
    )
    show.add_argument(
        '-t',
        default=None,
        type=str,
        help='run test after each holy value update (baw test -n1)',
    )



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import collections
import os
import re

import utilo

import configos
import configos.holyvalue.access
import configos.holyvalue.collect


def create(todo: list) -> dict:
    result = {}
    for path in todo:
        program = utilo.path_current(path)
        collected = configos.holyvalue.collect.collect(path)
        for groupname, group in collected.items():
            for key, value in group.items():
                progname = f'{program}.{groupname}'
                hvgroup = value.get('hvgroup', progname)
                if hvgroup == configos.holyvalue.access.NO_GROUP:
                    hvgroup = progname
                hvgroup = hvgroup.upper()
                variable = f'{hvgroup}.{key}'
                todo = ranges(
                    default=value.get('default'),
                    limit=value.get('limit', None),
                    datatype=value.get('datatype', None),
                )
                if not todo:
                    utilo.debug(f'skip for optimization, no range: {variable}')
                    continue
                result[variable] = todo
    return result


def run(
    path,
    reduce=100,
    seed=None,
    test_before: bool = False,
    cmd_test=None,
):
    path = os.path.abspath(path[0])
    plan = utilo.yaml_load(path)
    todo = list(plan.values())
    keys = list(plan.keys())
    mapped = first_one(todo)
    utilo.log(f'different steps: {len(mapped)}')
    if len(mapped) > reduce:
        utilo.log(f'reduce values: {reduce}')
        mapped = utilo.choose_random(mapped, count=reduce, seed=seed)
    # verify code without hv-modification
    if test_before:
        utilo.log('test project')
        utilo.run(cmd_test)  # utilo.run('baw test')
    # utilo.log(utilo.from_tuple(keys, ';'))
    with utilo.make_tmpdir(configos.ROOT) as tmpdir:
        utilo.log(f'outdir: {tmpdir}')
        header = f"number,{utilo.from_tuple(keys, separator=',')},failure\n"
        utilo.file_append(
            os.path.join(tmpdir, 'result'),
            header,
            create=True,
        )
        for index, step in enumerate(mapped):
            run_test(
                key=keys,
                config=step,
                step=index,
                tmpdir=tmpdir,
                cmd_test=cmd_test,
            )


def run_test(
    key,
    config,
    step: int,
    tmpdir,
    cmd_test,
    hcvalue='RAWMAKER',
):
    # write hv config
    cfg = create_config(key, config)
    step = str(step).zfill(4)
    cfgpath = os.path.join(tmpdir, f'{step}.hv')
    utilo.file_create(cfgpath, cfg)
    configos.cloud_set(program=hcvalue, namepath=cfgpath)
    # run tests
    utilo.log(f'run step: {step}')
    utilo.log(utilo.from_tuple(config, ';'))
    completed = utilo.run(
        cmd_test,
        expect=None,
        env=dict(os.environ),
    )
    # cfgpath
    logpath = os.path.join(tmpdir, f'{step}.log')
    utilo.file_create(logpath, completed.stderr)
    utilo.file_append(logpath, completed.stdout)
    tests = 0
    if completed.returncode:
        stdout = completed.stdout
        tests = FAILED.search(stdout)['failed']
    config = f'{step},' + utilo.from_tuple(config, ',') + f',{tests}\n'
    if completed.returncode:
        utilo.error(config)
    # append result
    utilo.file_append(
        os.path.join(tmpdir, 'result'),
        config,
        create=True,
    )


# === 5 failed, 178 passed, 168 skipped,
FAILED = re.compile(r'===\ (?P<failed>\d+)\ fail')

STEPS = (1.0, 0.1, 0.3, 0.5, 0.9, 1.2, 1.8, 2.2, 2.8, 3.5)


def ranges(
    default,
    limit=None,
    datatype: configos.DataType = None,
    steps=None,
) -> tuple:
    """\
    >>> ranges(50, 300, configos.DataType.INT_PLUS, steps=((1.0, 1.5, 2.0)))
    (50, 75, 100)
    >>> ranges(1.2, 3.0, configos.DataType.FLOAT, steps=((1.0, 1.5, 2.0)))
    (1.2, 1.8, 2.4)
    >>> ranges(True, datatype=configos.DataType.BOOL)
    (True, False)
    """
    if steps is None:
        steps = STEPS
    if default is None:
        return tuple()
    if datatype and datatype == configos.DataType.STR:
        return tuple()
    if datatype == configos.DataType.BOOL:
        return (True, False)
    data = tuple(utilo.roundme(default * item) for item in steps)
    if 'INT' in str(datatype):
        data = tuple(int(item) for item in data)
    if limit is not None:
        data = tuple(item for item in data if item < limit)
    return data


def create_config(keys, configs) -> str:
    grouped = collections.defaultdict(list)
    for key, config in zip(keys, configs):
        group, key = key.rsplit('.', 1)
        grouped[group].append(f'{key} = {config}')
    collected = []
    for key, value in grouped.items():
        collected.append('[' + key + ']')
        collected.extend(value)
    result = utilo.NEWLINE.join(collected)
    return result


def first_one(items) -> list:
    """\
    >>> first_one(([1,], [2,]))
    [[1, 2]]
    >>> first_one([[1, 2], [3,]])
    [[1, 3], [2, 3]]
    >>> first_one([(4, ), (1, 2, 3), (5, 6)])
    [[4, 1, 5], [4, 2, 5], [4, 3, 5], [4, 1, 6]]
    """
    # TODO: MOVE TO UTILA
    items = [list(item) for item in items]
    result = []
    for index, _ in enumerate(items):
        base = [item[0] for item in items]
        for current in items[index]:
            copy = list(base)
            copy[index] = current
            result.append(copy)
    result = utilo.unique(result)
    return result


def dump(plan: dict) -> str:
    dumped = utilo.yaml_dump(plan)
    return dumped



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import sys

import utilo

import configos
import configos.cli.generate
import configos.cli.optimization


@utilo.saveme
def main():
    current, data = evaluate()
    for action, method in runner():
        if current != action:
            continue
        if method.evaluate(*data):
            return utilo.FAILURE
        return utilo.SUCCESS
    return utilo.INVALID_COMMAND


def runner():
    runme = (
        ('generate', configos.cli.generate),
        ('optimize', configos.cli.optimization),
    )
    return runme


def evaluate() -> tuple:
    parser = create_parser()
    args = utilo.parse(parser)
    action, data = '', None
    gen = (args.get('input'), args.get('noskip', False))
    if args['generate']:
        action = 'generate'
        data = gen
    optimize = (
        args.get('create'),
        args.get('run'),
        args.get('show'),
        args.get('r'),
        args.get('t'),
    )
    if any(optimize):
        action = 'optimize'
        data = optimize
    if not action:
        utilo.error('nothing todo')
        sys.exit(utilo.INVALID_COMMAND)
    return action, data


def create_parser():
    result = utilo.cli.create_parser(
        todo=[
            utilo.cli.Flag(
                longcut='--generate',
                message='create default config out of source',
            ),
            utilo.cli.Flag(
                longcut='--noskip',
                message='do not skip any path',
            ),
        ],
        config=utilo.ParserConfiguration(
            inputparameter=True,
            outputparameter=False,
            cacheflag=False,
            cprofile=False,
            multiprocessed=False,
            pages=False,
            prefix=False,
            verboseflag=True,
            waitingflag=False,
        ),
        prog=configos.PROCESS,
        version=configos.__version__,
    )
    configos.cli.optimization.add_option(result)
    return result



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================


class HolyValueError(ValueError):
    pass


class InvalidHolyValue(HolyValueError):
    pass


class MissingHolyValue(HolyValueError):
    """Definine missing holy value name"""



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import os

import utilo

NO_DEFAULT = object()


@utilo.cacheme
def env(name: str, default=NO_DEFAULT, group: str = None):
    name = groupname(name, group)
    try:
        return os.environ[name]
    except KeyError as error:
        if default == NO_DEFAULT:
            raise error
    return default


def env_set(name: str, value: str, group: str = None):
    # TODO: MAKE THREAD SAFE?
    name = groupname(name, group)
    value = str(value)
    os.environ[name] = value
    utilo.cache_clear()


def env_del(name: str, group: str = None):
    name = groupname(name, group)
    del os.environ[name]
    utilo.cache_clear()


def env_load(path: str):
    assert os.path.exists(path), str(path)
    loaded = utilo.file_read(path)
    config = utilo.load_config(loaded, flat=True)
    for key, value in config.items():
        env_set(key, value)
    utilo.cache_clear()


def env_unload(path: str):
    assert os.path.exists(path), str(path)
    config = utilo.load_config(path, flat=True)
    for key in config.keys():
        env_del(key)
    utilo.cache_clear()


def env_dump() -> str:
    collected = []
    for key, value in os.environ.items():
        collected.append('{:<40}{}'.format(key, value))
    result = utilo.NEWLINE.join(collected)
    return result


def groupname(name, group):
    """\
    >>> groupname('first', 'master')
    'master_first'
    """
    if not group:
        return name
    return f'{group}_{name}'


def env_path_append(path: str):
    """\
    >>> env_path_append('/d/path/append')
    """
    base = f'{env("PATH")};{path}'
    env_set('PATH', base)
    utilo.debug(f'PATH:{base}')


def env_path_remove(path: str):
    """\
    >>> env_path_append('/d/path/not_there')
    >>> env_path_remove('/d/path/not_there')
    >>> env_path_remove('/d/path/not_there')
    Traceback (most recent call last):
    ...
    ValueError: path does not contain: /d/path/not_there
    """
    base = f'{env("PATH")}'
    remove = f';{path}'
    if remove not in base:
        raise ValueError(f'path does not contain: {path}')
    base = base.replace(remove, '')
    env_set('PATH', base)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import utilotest

import tests


def test_cli(mp):
    tests.run('--help', mp=mp)


def test_cli_generate(mp, capsys):
    source = tests.TEST_DATA
    tests.run(
        f'--generate -i {source} --noskip',
        mp=mp,
    )
    stdout = utilotest.stdout(capsys)
    assert 'HELMUT = None' in stdout
    assert len(stdout) >= 861


def test_cli_result_show(mp):
    result = tests.RESULT
    tests.run(
        f'optimize --show {result}',
        mp=mp,
    )


# @utilotest.longrun
# def test_cli_create_run_show(td, mp):
#     root = configos.ROOT
#     plan = utilo.forward_slash(str(td.tmpdir.join('plan.hv')))
#     with utilo.capture_stdout() as buffer:
#         tests.run(
#             f'optimize --create {root}',
#             mp=mp,
#         )
#     plan_content = buffer()
#     utilo.file_create(plan, plan_content)
#     utilo.run('baw init myproject "helm is here"')
#     with utilo.capture_stdout() as buffer:
#         tests.run(
#             f'optimize -r 2 --run {plan} ',
#             mp=mp,
#         )
#     # determine result path out of logging
#     result = buffer().strip()
#     path = utilo.search(r'outdir\:[ ](.+?)\n', result)[1]
#     path = utilo.join(path, 'result')
#     tests.run(
#         f'optimize --show {path}',
#         mp=mp,
#     )



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2020-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import os

DEFAULT_URL = 'http://www.checkitweg.de/writing/'


def docs_url() -> str:
    """Public writers club path.

    >>> docs_url()
    'http://.../writing/'
    """
    url = os.environ.get(
        'CAELUM_DOCS_MAIN',
        DEFAULT_URL,
    )
    assert url.endswith('/'), str(url)
    return url



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import pytest
import utilo

import configos

DUMP = """\
SHARED_SPACE                            /tmp/shared
HOSTNAME                                36049c61e901
SHLVL                                   2
HOME                                    /root
SHARED_READY                            /tmp/shared/ready
PATH                                    /opt/venv/bin:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin
BAW                                     /tmp/dev
SHARED_TODO                             /tmp/shared/todo
PYLINTHOME                              /tmp/pylint
PYTEST_PLUGINS                          pytester
PWD                                     /var/workdir
SHARED_TMP                              /tmp/shared/tmp
PYTEST_VERSION                          9.0.2
"""

DUMP_SIZE_MIN = len(DUMP)


def test_env_dump():
    dumped = configos.env_dump()
    assert len(dumped) > DUMP_SIZE_MIN


def test_env_get_set_del():
    with pytest.raises(KeyError):
        configos.env('abc')

    configos.env_set('abc', 10)
    assert configos.env('abc') == '10'

    configos.env_del('abc')
    with pytest.raises(KeyError):
        configos.env('abc')


CONFIG = """\
kiwi_rawmaker = 10
kiwi_rawmaker_asd = 15
ciwi_detector = Hier Spricht Helm
"""


def test_env_load_unload(td):
    config = td.tmpdir.join('config.ini')
    utilo.file_create(config, CONFIG)

    with pytest.raises(KeyError):
        configos.env('kiwi_rawmaker')

    configos.env_load(config)

    assert configos.env('kiwi_rawmaker') == '10'
    assert configos.env('kiwi_rawmaker_asd') == '15'
    assert configos.env('ciwi_detector') == 'Hier Spricht Helm'

    configos.env_unload(config)

    with pytest.raises(KeyError):
        configos.env('kiwi_rawmaker')


def test_env_unload(td):
    path = td.tmpdir.join('config.ini')
    utilo.file_create(path, CONFIG)
    before = configos.env_dump()
    configos.env_load(path)
    after = configos.env_dump()
    configos.env_unload(path)
    final = configos.env_dump()

    assert before != after
    assert final == before



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================
"""\
>>> @cache_large
... def hello():
...    pass
>>> hello()
>>> @cache_small
... def wello():
...    print('hello')
>>> wello()
hello
>>> wello()
>>> wello()
"""

import functools

import utilo

CACHE_SMALL = 32
CACHE_MEDIUM = 512
CACHE_LARGE = 4096

# TODO: MAY ONLY USE UTILA.CACHE???
cache_small = functools.partial(utilo.cacheme, maxsize=CACHE_SMALL)
cache_medium = functools.partial(utilo.cacheme, maxsize=CACHE_MEDIUM)
cache_large = functools.partial(utilo.cacheme, maxsize=CACHE_LARGE)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

# TODO: ENABLE LATER
# import os

# import utilotest

# import configos

# @utilotest.longrun
# def test_generate_rawmaker():
#     import rawmaker  # pylint:disable=C0415
#     source = os.path.join(rawmaker.ROOT, 'rawmaker')
#     generated = configos.generate(source)
#     assert '[rawmaker.' in generated
#     assert 'DIFF_MAX = [' in generated



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import os

import pytest

import configos
import configos.directory


def test_missing_environment(mp):
    """The communication folder for the view and pdf-mining is defined by
    shared folder `SHARED_TODO` and `SHARED_READY`"""
    with mp.context() as context:
        # Remove all environment vars
        context.setattr(os, 'environ', {})

        with pytest.raises(SystemExit):
            configos.check_startup()

        with pytest.raises(SystemExit):
            configos.todo()

        with pytest.raises(SystemExit):
            configos.ready()


def test_wrong_todo(mp):
    with mp.context() as context:
        context.setattr(
            os, 'environ', {
                configos.directory.TODO: 'ThisPathDoesNotExist',
                configos.directory.READY: 'ThisPathDoesNotExist',
                configos.directory.COMMON: 'ThisPathDoesNotExist',
            })

        with pytest.raises(SystemExit):
            configos.todo(check=True)

        with pytest.raises(SystemExit):
            configos.ready(check=True)


def test_startup(td, mp):
    root = str(td)
    ready = os.path.join(root, 'ready')
    tmp = os.path.join(root, 'tmp')
    todo = os.path.join(root, 'todo')
    for item in (ready, tmp, todo):
        os.makedirs(item)

    with mp.context() as context:
        context.setattr(
            os, 'environ', {
                configos.directory.TODO: todo,
                configos.directory.READY: ready,
                configos.directory.TMP: tmp,
                configos.directory.COMMON: root
            })
        configos.check_startup()


def test_export_import(tmpdir, mp):
    with mp.context() as context:
        # Remove all environment vars
        context.setattr(os, 'environ', {})

        todo = os.path.join(tmpdir, 'todo')  #pylint:disable=W0621
        ready = os.path.join(tmpdir, 'ready')  #pylint:disable=W0621
        common = os.path.join(tmpdir, 'common')

        for item in (todo, ready, common):
            os.makedirs(item)

        configos.export(common, todo, ready)
        common_, todo_, ready_, = configos.environment(True)

        assert todo_ == todo
        assert ready_ == ready
        assert common_ == common


def test_without_folder_configuration(mp):

    with mp.context() as context:
        # Remove all environment vars
        context.setattr(os, 'environ', {})

        with pytest.raises(SystemExit):
            configos.environment()


def test_with_wrong_folder_configuration(mp):
    """The communication folder for the view and pdf-mining is defined by
    shared folder `SHARED_TODO` and `SHARED_READY`"""

    with mp.context() as context:
        # Remove all environment vars
        context.setattr(
            os, 'environ', {
                'SHARED_SPACE': 'NO_PATH',
                'SHARED_TODO': 'NO_PATH',
                'SHARED_READY': 'NO_PATH',
            })

        with pytest.raises(SystemExit):
            configos.environment(check=True)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import utilo

import configos.cli.plan
import tests.hvexample


def test_plan_create():
    """Ensure that hvexample is loaded to ensure that hv is loaded correctly."""
    path = utilo.path_parent(tests.hvexample.__file__)
    todo = [path]
    plan = configos.cli.plan.create(todo)
    assert 'TESTS.HVEXAMPLE.SECOND' in plan



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================
"""\
Environment variables:

>>> HC_RAWMAKER = 'rawmaker'
>>> HC_WORDS = '/c/holyclound/words341.hv'
"""

import utilo

import configos

HC_CLOUD_BASE = 'HC_CLOUD_BASE'


def cloud_lookup(program: str, base: str = None):
    envname = holyname(program)
    env = configos.env(envname, default=None)
    if env is None:
        return
    if base:
        configos.init(base)
    if utilo.isfilepath(env):
        if not utilo.exists(env):
            utilo.error(f'invalid holy value path[env:{envname}]: {env}')
            return
        base = utilo.path_parent(env)
        name = utilo.file_name(env)
        # hv-file path
        configos.load(name=name, base=base)
        return
    # hv-file name
    configos.load(name=env)


def cloud_set(program: str, namepath: str = None):
    if not namepath:
        namepath = program
    program = holyname(program)
    configos.env_set(program, value=namepath)


def cloud_unset(program: str):
    program = holyname(program)
    configos.env_del(program)


def cloud_base_set(path: str):
    if path is None:
        configos.env_del(HC_CLOUD_BASE)
        return
    configos.env_set(HC_CLOUD_BASE, path)


def cloud_base() -> str:
    result = configos.env(HC_CLOUD_BASE, default=None)
    return result


def holyname(program: str) -> str:
    """\
    >>> holyname('rawmaker')
    'HC_RAWMAKER'
    """
    assert program
    program = program.upper()
    return f'HC_{program}'



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import pytest

import configos


def test_second_fine():
    _ = configos.HV_SECOND(default=32)


def test_second_failure():
    with pytest.raises(configos.InvalidHolyValue):
        _ = configos.HV_SECOND(default=-32)
    with pytest.raises(configos.InvalidHolyValue):
        _ = configos.HV_SECOND(default=16.0)


def test_api_prefix():
    _ = configos.HV_API(default='/api/v0')


def test_secret_key():
    value = 'iamimportant'
    secret = configos.HV_SECRET(default=value)
    assert isinstance(secret.value, bytes)


def test_xb():
    kb = configos.HV_KB(default=25)  # pylint:disable=C0103
    assert kb == 25 * 1024
    mb = configos.HV_MB(default=25)  # pylint:disable=C0103
    assert mb == 25 * 1024 * 1024
    gb = configos.HV_GB(default=25)  # pylint:disable=C0103
    assert gb == 25 * 1024 * 1024 * 1024



#!/usr/bin/env python
#==============================================================================
# C O P Y R I G H T
#------------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
#==============================================================================

import os

import utilo

PACKAGES = [
    'configos',
    'configos.cli',
    'configos.holyvalue',
]
ENTRY_POINTS = {
    'console_scripts': ['configos = configos.cli:main'],
}
if __name__ == "__main__":
    utilo.install(__file__)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import pytest
import utilo

import configos
import tests


def test_cloud_config(td):
    program = 'markers'
    source = tests.TEST_DATA
    holyconfig = td.tmpdir.join('config.hv')
    utilo.run(f'configos --generate -i {source} --noskip >> config.hv')
    configos.cloud_set(program, holyconfig)
    configos.cloud_lookup(program)
    holypath = configos.env(configos.holyname(program))
    assert holypath
    assert holypath == holyconfig
    # unload config
    configos.cloud_unset(program)
    with pytest.raises(KeyError):
        # config was unloaded
        configos.env(configos.holyname(program))



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import importlib.util

import utilo


def load_module(path: str):
    """\
    >>> load_module(__file__).__name__
    'configos.utils'
    """
    item = utilo.file_name(path)
    parent = utilo.file_name(utilo.path_parent(path))
    spec = importlib.util.spec_from_file_location(
        f'{parent}.{item}',
        path,
    )
    try:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except Exception:  # pylint:disable=broad-except
        return None
    return module



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

import utilo

import configos


def evaluate(inpath: list, noskip=False) -> int:
    """\
    Could not locate any holy value on this file. 1 means failure.
    >>> evaluate([__file__])
    1
    """
    skip = None if noskip else skips
    for path in inpath:
        if not utilo.exists(path):
            utilo.error(f'input does not exists: {path}')
            return utilo.FAILURE
    done = False
    for item in inpath:
        collected = configos.generate(item, skips=skip)
        if not collected:
            continue
        utilo.print_banner(text=item, symbol='#')
        utilo.log(collected)
        done = True
    if not done:
        utilo.error('could not locate any HolyValue')
        return utilo.FAILURE
    return utilo.SUCCESS


def skips(item: str) -> bool:
    """\
    >>> skips('config/tests/__init.py')
    True
    """
    item = str(item)
    return 'build' in item or 'tests' in item



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================

from configos import OneSideDINA4
from configos import OneSideDINA5


def test_configuration_access():
    border = OneSideDINA4.pageborder
    left, right, top, bottom = border  # pylint: disable=unused-variable


def test_different_pagesize():
    assert len(OneSideDINA4.pagesize) == len(OneSideDINA5.pagesize)
    assert len(OneSideDINA4.pagesize) == 2
    assert OneSideDINA4.pagesize != OneSideDINA5.pagesize

    assert len(OneSideDINA4.pageborder) == len(OneSideDINA5.pageborder)



# =============================================================================
# C O P Y R I G H T
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2022 by Helmut Konrad Fahrendholz. All rights reserved.
# This file is property of Helmut Konrad Fahrendholz. Any unauthorized copy,
# use or distribution is an offensive act against international law and may
# be prosecuted under federal law. Its content is company confidential.
# =============================================================================
"""Document Constants
==================

This class provides information about the `expected` parameter of the
layout-checker-configuration.

For the first approach, DinA4 is enough, DinA5 is included to keep supporting
different page formats and pageconfiguration(oneside/twoside) in mind.
"""

import collections
import dataclasses

PageSize = collections.namedtuple('PageSize', 'width height')
PageBorder = collections.namedtuple('PageBorder', 'left right top bottom')

DINA4 = PageSize(210, 297)
DINA5 = PageSize(148, 210)

# Manuel Rene Theisen, Wissenschaftliches Arbeiten: Erfolgreich bei Bachelor-
# und Masterarbeit.
BORDER_DINA4 = PageBorder(40, 20, 40, 20)

# TODO: Make configuration readonly?


@dataclasses.dataclass
class DocumentConfiguration:
    """Base class of all document configuration"""
    pagesize: PageSize = None
    pageborder: PageBorder = None


class OneSideDINA4(DocumentConfiguration):
    pagesize = DINA4
    pageborder = BORDER_DINA4


class OneSideDINA5(DocumentConfiguration):
    pagesize = DINA5
    pageborder = BORDER_DINA4  # What is a good value for DINA5?
